/*
 * Wired Ethernet driver interface
 * Copyright (c) 2005-2009, Jouni Malinen <j@w1.fi>
 * Copyright (c) 2004, Gunter Burchardt <tira@isx.de>
 *
 * This software may be distributed under the terms of the BSD license.
 * See README for more details.
 */

#include <sys/stat.h>
#include "includes.h"

#include "common.h"
#include "eloop.h"
#include "driver.h"
#include "netlink.h"
#include "ap/sta_info.h"
#include "ap/hostapd.h"
#include "ap/ap_drv_ops.h"
#include "common/ieee802_11_defs.h"
#include "driver_wired_common.h"
#include "eapol_auth/eapol_auth_sm.h"
#include "eapol_auth/eapol_auth_sm_i.h"
#include "eap_server/eap.h"

#include <sys/ioctl.h>
#include <linux/filter.h>
#include <linux/rtnetlink.h>
#include <netlink/netlink.h>
#include <netlink/cache.h>
#include <netlink/route/link/bridge.h>
#undef IFNAMSIZ
#include <net/if.h>
#ifdef __linux__
#include <netpacket/packet.h>
#include <net/if_arp.h>
#endif /* __linux__ */
#if defined(__FreeBSD__) || defined(__DragonFly__) || defined(__FreeBSD_kernel__)
#include <net/if_dl.h>
#include <net/if_media.h>
#endif /* defined(__FreeBSD__) || defined(__DragonFly__) || defined(__FreeBSD_kernel__) */
#ifdef __sun__
#include <sys/sockio.h>
#endif /* __sun__ */

#ifdef _MSC_VER
#pragma pack(push, 1)
#endif /* _MSC_VER */

#ifndef IFF_LOWER_UP
#define IFF_LOWER_UP   0x10000         /* driver signals L1 up         */
#endif
#ifndef IFF_DORMANT
#define IFF_DORMANT    0x20000         /* driver signals dormant       */
#endif

#define DROP_PRIORITY 50
const char aclcommand[] = "/usr/cumulus/bin/cl-acltool";
const char aclargs[] = "-i";
const char tccommand[] = "/sbin/tc";
const char bridge_command[] = "/sbin/bridge";

struct ieee8023_hdr {
	u8 dest[6];
	u8 src[6];
	u16 ethertype;
} STRUCT_PACKED;

struct ether_vlan_hdr {
	u8 dest[6];
	u8 src[6];
	u16 tpid;
	u16 tci;
} STRUCT_PACKED;

#ifdef _MSC_VER
#pragma pack(pop)
#endif /* _MSC_VER */


struct wired_global {
	void *ctx;
	struct dl_list interfaces;
	struct netlink_data *netlink;
	struct nl_cache_mngr *mngr;
	struct nl_cache *cache;
	struct nl_sock *sync_sock;
};

struct wpa_driver_wired_if_info {
	int pvid;
	int bridge_id;
	int is_bridge;
	u32 flags;
};

#define WPA_WIRED_DATA_PVLAN_ENABLED BIT(0)
#define WPA_WIRED_DATA_DYNAMIC_VLAN_ENABLED BIT(1)

struct wpa_driver_wired_data {
	struct dl_list  if_list;
	struct wired_global *global;
	struct driver_wired_common_data common;

	int dhcp_sock; /* socket for dhcp packets */
	int mab_sock; /* socket for learning source MACs */
	int use_pae_group_addr;
	int flags;
	struct wpa_driver_wired_if_info if_info;
};


/* TODO: detecting new devices should eventually be changed from using DHCP
 * snooping to trigger on any packet from a new layer 2 MAC address, e.g.,
 * based on ebtables, etc. */

struct dhcp_message {
	u_int8_t op;
	u_int8_t htype;
	u_int8_t hlen;
	u_int8_t hops;
	u_int32_t xid;
	u_int16_t secs;
	u_int16_t flags;
	u_int32_t ciaddr;
	u_int32_t yiaddr;
	u_int32_t siaddr;
	u_int32_t giaddr;
	u_int8_t chaddr[16];
	u_int8_t sname[64];
	u_int8_t file[128];
	u_int32_t cookie;
	u_int8_t options[308]; /* 312 - cookie */
};

/* Berkeley Packet filter code to filter out BFD Echo packets.
 * tcpdump -dd -n inbound"
 */
static struct sock_filter mab_sock_filter[] = {
	{ 0x28, 0, 0, 0xfffff004 },
	{ 0x15, 0, 1, 0x00000004 },
	{ 0x6, 0, 0, 0x00000000 },
	{ 0x6, 0, 0, 0x00040000 },
};

void wired_mab_deactivation(struct hostapd_data *hapd);

static int
wired_driver_hapd_is_l2_if(void *priv)
{
	struct wpa_driver_wired_data *drv = priv;
	return drv->if_info.is_bridge;
}

static int
wired_flush(void *priv)
{
	struct wpa_driver_wired_data *drv = priv;
	struct hostapd_data *hapd;
	struct sta_info *sta;

	wpa_printf(MSG_DEBUG, "WIRED: wired_flush called");
	hapd = drv->common.ctx;
	if (!hapd)
		return 0;
	for (sta = hapd->sta_list; sta; sta = sta->next) {
		hostapd_drv_sta_deauth(hapd, sta->addr,
					WLAN_REASON_DISASSOC_STA_HAS_LEFT);
		ap_sta_deauthenticate(hapd, sta, WLAN_REASON_DISASSOC_STA_HAS_LEFT);
	}
	hostapd_free_stas(hapd);
	return 0;
}

#ifdef __linux__
static void handle_data(void *ctx, unsigned char *buf, size_t len)
{
#ifdef HOSTAPD
	struct ieee8023_hdr *hdr;
	u8 *pos, *sa;
	struct hostapd_data *hapd = ctx;
	size_t left;
	union wpa_event_data event;

	wpa_printf(MSG_ERROR, "Handle_data:recv: packet on %s", hapd->conf->iface);

	/* must contain at least ieee8023_hdr 6 byte source, 6 byte dest,
	 * 2 byte ethertype */
	if (len < 14) {
		wpa_printf(MSG_MSGDUMP, "handle_data: too short (%lu)",
			   (unsigned long) len);
		return;
	}

	hdr = (struct ieee8023_hdr *) buf;
	sa = hdr->src;
	if (os_memcmp(sa, hapd->own_addr, ETH_ALEN) == 0)
		return;

	if (ntohs(hdr->ethertype) == ETH_P_PAE) {
		wpa_printf(MSG_MSGDUMP, "Received EAPOL packet");
		hapd->conf->pae_detected = TRUE;
		wired_mab_deactivation(hapd);
	} else {
		wpa_printf(MSG_MSGDUMP, "Received NON-EAPOL packet");
		hapd->conf->pae_detected = FALSE;
	}

	os_memset(&event, 0, sizeof(event));
	event.new_sta.addr = sa;
	wpa_supplicant_event(ctx, EVENT_NEW_STA, &event);

	if (ntohs(hdr->ethertype) == ETH_P_PAE) {
		pos = (u8 *) (hdr + 1);
		left = len - sizeof(*hdr);
		drv_event_eapol_rx(ctx, sa, pos, left);
	}

#endif /* HOSTAPD */
}

static void handle_read(int sock, void *eloop_ctx, void *sock_ctx)
{
	int len;
	unsigned char buf[3000];

	len = recv(sock, buf, sizeof(buf), 0);
	if (len < 0) {
		wpa_printf(MSG_ERROR, "recv: %s", strerror(errno));
		return;
	}

	handle_data(eloop_ctx, buf, len);
}

struct rtnl_link *wpa_nl_get_link_by_family(struct nl_cache *cache,
											int family, int ifindex)
{
	struct rtnl_link *link, *link_filter;

	if (!(link_filter = rtnl_link_alloc()))
		return NULL;

	rtnl_link_set_family(link_filter, family);
	rtnl_link_set_ifindex(link_filter, ifindex);

	link = (struct rtnl_link *)nl_cache_search(cache, OBJ_CAST(link_filter));

	rtnl_link_put(link_filter);

	return link;
}

int wpa_nl_is_link_bridge(struct rtnl_link *link)
{
	char *type = rtnl_link_get_type(link);

	if (type)
		wpa_printf(MSG_DEBUG, "%s: type not found", __FUNCTION__);
	else
		wpa_printf(MSG_DEBUG, "%s: type %s", __FUNCTION__, type);
	return type && !strcmp(type, "bridge");
}

static void handle_dhcp(int sock, void *eloop_ctx, void *sock_ctx)
{
	int len;
	unsigned char buf[3000];
	struct dhcp_message *msg;
	u8 *mac_address;
	union wpa_event_data event;

	len = recv(sock, buf, sizeof(buf), 0);
	if (len < 0) {
		wpa_printf(MSG_ERROR, "recv: %s", strerror(errno));
		return;
	}

	/* must contain at least dhcp_message->chaddr */
	if (len < 44) {
		wpa_printf(MSG_MSGDUMP, "handle_dhcp: too short (%d)", len);
		return;
	}

	msg = (struct dhcp_message *) buf;
	mac_address = (u8 *) &(msg->chaddr);

	wpa_printf(MSG_MSGDUMP, "Got DHCP broadcast packet from " MACSTR,
		   MAC2STR(mac_address));

	os_memset(&event, 0, sizeof(event));
	event.new_sta.addr = mac_address;
	wpa_supplicant_event(eloop_ctx, EVENT_NEW_STA, &event);
}
#endif /* __linux__ */


static int wired_init_sockets(struct wpa_driver_wired_data *drv, u8 *own_addr)
{
#ifdef __linux__
	struct ifreq ifr;
	struct sockaddr_ll addr;
	struct sockaddr_in addr2;
	int n = 1;

	/* XXXXX FIXME change to IPv4 for testing */
	drv->common.sock = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_PAE));
	if (drv->common.sock < 0) {
		wpa_printf(MSG_ERROR, "socket[PF_PACKET,SOCK_RAW]: %s",
			   strerror(errno));
		return -1;
	}

	if (eloop_register_read_sock(drv->common.sock, handle_read,
				     drv->common.ctx, NULL)) {
		wpa_printf(MSG_INFO, "Could not register read socket");
		return -1;
	}

	os_memset(&ifr, 0, sizeof(ifr));
	os_strlcpy(ifr.ifr_name, drv->common.ifname, sizeof(ifr.ifr_name));
	if (ioctl(drv->common.sock, SIOCGIFINDEX, &ifr) != 0) {
		wpa_printf(MSG_ERROR, "ioctl(SIOCGIFINDEX): %s",
			   strerror(errno));
		return -1;
	}

	os_memset(&addr, 0, sizeof(addr));
	addr.sll_family = AF_PACKET;
	addr.sll_ifindex = ifr.ifr_ifindex;
	drv->common.ifindex = ifr.ifr_ifindex;
	wpa_printf(MSG_DEBUG, "Opening raw packet socket for ifindex %d",
		   addr.sll_ifindex);

	if (bind(drv->common.sock, (struct sockaddr *) &addr, sizeof(addr)) < 0)
	{
		wpa_printf(MSG_ERROR, "bind: %s", strerror(errno));
		return -1;
	}

	/* filter multicast address */
	if (wired_multicast_membership(drv->common.sock, ifr.ifr_ifindex,
				       pae_group_addr, 1) < 0) {
		wpa_printf(MSG_ERROR, "wired: Failed to add multicast group "
			   "membership");
		return -1;
	}

	os_memset(&ifr, 0, sizeof(ifr));
	os_strlcpy(ifr.ifr_name, drv->common.ifname, sizeof(ifr.ifr_name));
	if (ioctl(drv->common.sock, SIOCGIFHWADDR, &ifr) != 0) {
		wpa_printf(MSG_ERROR, "ioctl(SIOCGIFHWADDR): %s",
			   strerror(errno));
		return -1;
	}

	if (ifr.ifr_hwaddr.sa_family != ARPHRD_ETHER) {
		wpa_printf(MSG_INFO, "Invalid HW-addr family 0x%04x",
			   ifr.ifr_hwaddr.sa_family);
		return -1;
	}
	os_memcpy(own_addr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);

	/* setup dhcp listen socket for sta detection */
	if ((drv->dhcp_sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
		wpa_printf(MSG_ERROR, "socket call failed for dhcp: %s",
			   strerror(errno));
		return -1;
	}

	if (eloop_register_read_sock(drv->dhcp_sock, handle_dhcp,
				     drv->common.ctx, NULL)) {
		wpa_printf(MSG_INFO, "Could not register read socket");
		return -1;
	}

	os_memset(&addr2, 0, sizeof(addr2));
	addr2.sin_family = AF_INET;
	addr2.sin_port = htons(67);
	addr2.sin_addr.s_addr = INADDR_ANY;

	if (setsockopt(drv->dhcp_sock, SOL_SOCKET, SO_REUSEADDR, (char *) &n,
		       sizeof(n)) == -1) {
		wpa_printf(MSG_ERROR, "setsockopt[SOL_SOCKET,SO_REUSEADDR]: %s",
			   strerror(errno));
		return -1;
	}
	if (setsockopt(drv->dhcp_sock, SOL_SOCKET, SO_BROADCAST, (char *) &n,
		       sizeof(n)) == -1) {
		wpa_printf(MSG_ERROR, "setsockopt[SOL_SOCKET,SO_BROADCAST]: %s",
			   strerror(errno));
		return -1;
	}

	os_memset(&ifr, 0, sizeof(ifr));
	os_strlcpy(ifr.ifr_ifrn.ifrn_name, drv->common.ifname, IFNAMSIZ);
	if (setsockopt(drv->dhcp_sock, SOL_SOCKET, SO_BINDTODEVICE,
		       (char *) &ifr, sizeof(ifr)) < 0) {
		wpa_printf(MSG_ERROR,
			   "setsockopt[SOL_SOCKET,SO_BINDTODEVICE]: %s",
			   strerror(errno));
		return -1;
	}

	if (bind(drv->dhcp_sock, (struct sockaddr *) &addr2,
		 sizeof(struct sockaddr)) == -1) {
		wpa_printf(MSG_ERROR, "bind: %s", strerror(errno));
		return -1;
	}

	return 0;
#else /* __linux__ */
	return -1;
#endif /* __linux__ */
}


static int wired_send_eapol(void *priv, const u8 *addr,
			    const u8 *data, size_t data_len, int encrypt,
			    const u8 *own_addr, u32 flags)
{
	struct wpa_driver_wired_data *drv = priv;
	struct ieee8023_hdr *hdr;
	size_t len;
	u8 *pos;
	int res;

	len = sizeof(*hdr) + data_len;
	hdr = os_zalloc(len);
	if (hdr == NULL) {
		wpa_printf(MSG_INFO,
			   "malloc() failed for wired_send_eapol(len=%lu)",
			   (unsigned long) len);
		return -1;
	}

	os_memcpy(hdr->dest, drv->use_pae_group_addr ? pae_group_addr : addr,
		  ETH_ALEN);
	os_memcpy(hdr->src, own_addr, ETH_ALEN);
	hdr->ethertype = htons(ETH_P_PAE);

	pos = (u8 *) (hdr + 1);
	os_memcpy(pos, data, data_len);

	res = send(drv->common.sock, (u8 *) hdr, len, 0);
	os_free(hdr);

	if (res < 0) {
		wpa_printf(MSG_ERROR,
			   "wired_send_eapol - packet len: %lu - failed: send: %s",
			   (unsigned long) len, strerror(errno));
	}

	return res;
}

void wired_delete_tc_qdisc (char *ifname)
{
	char tcargs[128];

	sprintf(tcargs, "qdisc del dev %s handle ffff: ingress", ifname);
	if (os_exec(tccommand, tcargs, 1)) {
		wpa_printf(MSG_DEBUG, "WIRED: could not delete ingress tc");
	}

	sprintf(tcargs, "qdisc del dev %s root handle 1: prio", ifname);
	if (os_exec(tccommand, tcargs, 1)) {
		wpa_printf(MSG_ERROR, "WIRED: could not delete egress tc");
	}
}

#define MAC4BYTE(a) (a)[2], (a)[3], (a)[4], (a)[5]
#define COMPACT_MAC4BYTE "%02x%02x%02x%02x"
#define MAC2BYTE(a) (a)[0], (a)[1]
#define COMPACT_MAC2BYTE "%02x%02x"

static int wired_tc_exec (FILE *tc_f, char *tcargs) {
	fprintf(tc_f, "%s\n", tcargs);
	if (os_exec(tccommand, tcargs, 1)) {
	    wpa_printf(MSG_ERROR, "WIRED: Error: could not run tc");
	    return -1;
	}

	return 0;
}

static int wired_handle_ebtable_rules(struct wpa_driver_wired_data *drv,
					struct sta_info *changed_sta,
					Boolean auth)
{
	struct hostapd_data *hapd;
	struct sta_info *sta;
	char rulefile[128];
	char tcrulefile[128];
	char tcargs[512];
	int i = 0;
	int starting_prio = 2;
	struct stat st = {0};
	FILE *tc_f = NULL;
	FILE *f = NULL;
	int delete_tc = 0;
	int mac_whitelist = 0;

	hapd = drv->common.ctx;
	wpa_printf(MSG_DEBUG, "WIRED: wired_handle_ebtable_rules called");

	/* rewrite the rule file for this interface */
	sprintf(rulefile, "/etc/cumulus/acl/policy.d/100_dot1x_%s.rules",
			drv->common.ifname);
	sprintf(tcrulefile, "/var/lib/hostapd/acl/tc_%s.rules", drv->common.ifname);

	if (stat(tcrulefile, &st) != -1)
		delete_tc = 1;

	tc_f = fopen(tcrulefile, "w");
	if (!tc_f) {
		wpa_printf(MSG_ERROR, "WIRED: ERROR: could not write tc rule file %s",
				tcrulefile);
		return -1;
	}

	f = fopen(rulefile, "w");
	if (!f) {
		wpa_printf(MSG_ERROR, "WIRED: ERROR: could not write ebtables rule"
				" file %s", rulefile);
		return -1;
	}

	fprintf(f, "# Generated 802.1x EAPOL EBTABLES Rules File: Do not edit.\n");
	fprintf(f, "[ebtables]\n");

	fprintf(tc_f, "# Generated 802.1x EAPOL TC Rules File: Do not edit.\n");

	if (delete_tc) {
		wired_delete_tc_qdisc (drv->common.ifname);
	}

	sprintf(tcargs, "qdisc add dev %s handle ffff: ingress", drv->common.ifname);
	if (wired_tc_exec(tc_f, tcargs) < 0)
		return -1;

	sprintf(tcargs, "qdisc add dev %s root handle 1: prio", drv->common.ifname);
	if (wired_tc_exec(tc_f, tcargs) < 0)
		return -1;

	sprintf(tcargs, "filter add dev %s parent 1: protocol 0x888E prio %d"
			" flower flowid :1 action ok",
			drv->common.ifname, starting_prio + ++i);
	if (wired_tc_exec(tc_f, tcargs) < 0)
		return -1;

	for (sta = hapd->sta_list; sta; sta = sta->next) {
		wpa_printf(MSG_DEBUG, "WIRED: sta addr " MACSTR " flags=%x",
				MAC2STR(sta->addr), sta->flags);
		if (!auth && sta == changed_sta)
			continue;

		/* we only allow traffic for authorized macs */
		if ((sta->flags & WLAN_STA_AUTHORIZED) ||
			(auth && (sta == changed_sta))) {
			mac_whitelist = 1;
			fprintf(f, "-A INPUT,FORWARD -i %s -s " MACSTR " -j mark --set-mark 2\n",
					drv->common.ifname, MAC2STR(sta->addr));
			fprintf(f, "-A INPUT,FORWARD -i %s -s " MACSTR " -j ACCEPT\n",
					drv->common.ifname, MAC2STR(sta->addr));

			sprintf(tcargs, "filter add dev %s parent 1: prio %d u32 match u32 0x"
					COMPACT_MAC4BYTE " 0xFFFFFFFF at -12 match u16 0x"
					COMPACT_MAC2BYTE " 0xFFFF at -14 flowid :1 action pass",
					drv->common.ifname,  ++i + starting_prio,
					MAC4BYTE(sta->addr), MAC2BYTE(sta->addr));
			if (wired_tc_exec(tc_f, tcargs) < 0)
				return -1;
		}
	}

	if (mac_whitelist) {
		sprintf(tcargs, "filter add dev %s parent 1: prio %d u32 match u8"
				" 0x01 0x01 at -14 flowid :1 action pass",
				drv->common.ifname, ++i + starting_prio);
		if (wired_tc_exec(tc_f, tcargs) < 0)
			return -1;
	}

	sprintf(tcargs, "filter add dev %s parent 1: prio %d matchall flowid :1 action drop",
		drv->common.ifname, DROP_PRIORITY);
	if (wired_tc_exec(tc_f, tcargs) < 0)
		return -1;

	fprintf(f, "-A INPUT -i %s -p 0x888E -j mark --set-mark 2\n", drv->common.ifname);
	fprintf(f, "-A INPUT -i %s -p 0x888E -j police --set-mode pkt --set-rate 100 --set-burst 100\n", drv->common.ifname);
	fprintf(f, "-A INPUT,FORWARD -i %s -j mark --set-mark 2\n", drv->common.ifname);
	fprintf(f, "-A INPUT,FORWARD -i %s -j DROP\n", drv->common.ifname);

	fclose(f);
	fclose(tc_f);

	/* Only call cl-acltool on the rules files if we are not starting or stopping */
	if (!hapd->iface->interfaces->start_in_progress && !hapd->iface->interfaces->stop_in_progress) {
		if (os_exec(aclcommand, aclargs, 1)) {
			wpa_printf(MSG_ERROR, "WIRED: Error: could not run cl-acltool");
			return -1;
		}
	}

	return 0;
}
 
static int wired_start_mab(struct wpa_driver_wired_data *drv)
{
	char rulefile[128];
	char tcrulefile[128];
	char tcargs[256];
	int starting_prio = 2;
	struct stat st = {0};
	FILE *tc_f = NULL;
	FILE *f = NULL;
	int delete_tc = 0;
	struct sockaddr_ll addr;
	struct hostapd_data *hapd;
	struct sock_fprog bpf =
	{
		.len = sizeof(mab_sock_filter) / sizeof (mab_sock_filter[0]),
		.filter = mab_sock_filter
	};

	wpa_printf(MSG_DEBUG, "wired_start_mab called:");
	hapd = drv->common.ctx;
	if (drv->mab_sock)
		return 0;

	drv->mab_sock = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
	if (drv->mab_sock < 0) {
		wpa_printf(MSG_ERROR, "socket[PF_PACKET,SOCK_RAW,ETH_P_ALL]: %s",
					strerror(errno));
		return -1;
	}
	wpa_printf(MSG_DEBUG, "wired_start_mab: created new socket");

	if (eloop_register_read_sock(drv->mab_sock, handle_read,
				     drv->common.ctx, NULL)) {
		wpa_printf(MSG_INFO, "Could not register read socket");
		return -1;
	}

	os_memset(&addr, 0, sizeof(addr));
	addr.sll_family = AF_PACKET;
	addr.sll_ifindex = drv->common.ifindex;
	wpa_printf(MSG_DEBUG, "Opening raw packet socket for ifindex %d",
				addr.sll_ifindex);

	if (bind(drv->mab_sock, (struct sockaddr *) &addr, sizeof(addr)) < 0)
	{
		wpa_printf(MSG_ERROR, "bind: %s", strerror(errno));
		return -1;
	}

	if (setsockopt(drv->mab_sock, SOL_SOCKET, SO_ATTACH_FILTER, &bpf,
			sizeof(bpf)) < 0) {
		wpa_printf(MSG_ERROR, "Setting mab socket filter for %d failed",
				addr.sll_ifindex);
		return-1;
	}

	/* rewrite the rule file for this interface */
	sprintf(rulefile, "/etc/cumulus/acl/policy.d/100_dot1x_%s.rules",
				drv->common.ifname);
	sprintf(tcrulefile, "/var/lib/hostapd/acl/tc_%s.rules",
				drv->common.ifname);

	if (stat(tcrulefile, &st) != -1)
		delete_tc = 1;
	tc_f = fopen(tcrulefile, "w");
	if (!tc_f) {
		wpa_printf(MSG_ERROR, "WIRED: ERROR: could not write tc rule file %s",
					tcrulefile);
		return -1;
	}

	f = fopen(rulefile, "w");
	if (!f) {
		wpa_printf(MSG_ERROR, "WIRED: ERROR: could not write ebtables rule"
					" file %s", rulefile);
		return -1;
	}

	fprintf(f, "# Generated 802.1x EAPOL EBTABLES Rules File: Do not edit.\n");
	fprintf(f, "[ebtables]\n");

	if (delete_tc) {
		wired_delete_tc_qdisc (drv->common.ifname);
	}

	fprintf(tc_f, "# Generated 802.1x EAPOL TC Rules File: Do not edit.\n");
	sprintf(tcargs, "qdisc add dev %s handle ffff: ingress", drv->common.ifname);
	if (wired_tc_exec(tc_f, tcargs) < 0)
		return -1;

	sprintf(tcargs, "filter add dev %s parent ffff: prio %d matchall flowid :1 action drop",
			drv->common.ifname, DROP_PRIORITY);
	if (wired_tc_exec(tc_f, tcargs) < 0)
		return -1;

	sprintf(tcargs, "qdisc add dev %s root handle 1: prio", drv->common.ifname);
	if (wired_tc_exec(tc_f, tcargs) < 0)
		return -1;

	sprintf(tcargs, "filter add dev %s parent 1: prio %d matchall flowid :1 action drop",
			drv->common.ifname, DROP_PRIORITY);
	if (wired_tc_exec(tc_f, tcargs) < 0)
		return -1;

	sprintf(tcargs, "filter add dev %s parent 1: protocol 0x888E prio %d"
				" flower flowid :1 action ok",
				drv->common.ifname, starting_prio + 1);
	if (wired_tc_exec(tc_f, tcargs) < 0)
		return -1;

	fprintf(f, "-A FORWARD -i %s -j mark --set-mark 3\n", drv->common.ifname);
	fprintf(f, "-A FORWARD -i %s -j DROP\n", drv->common.ifname);

	fclose(f);
	fclose(tc_f);

	/* call cl-acltool on the rules files */
	if (!hapd->iface->interfaces->start_in_progress && !hapd->iface->interfaces->stop_in_progress) {
		if (os_exec(aclcommand, aclargs, 1)) {
			wpa_printf(MSG_ERROR, "WIRED: Error: could not run cl-acltool");
			return -1;
		}
	}

	return 0;
}

static void wired_stop_mab(struct wpa_driver_wired_data *drv)
{
	if (drv->mab_sock > 0) {
		eloop_unregister_read_sock(drv->mab_sock);
		close(drv->mab_sock);
		drv->mab_sock = 0;
		wpa_printf(MSG_DEBUG, "wired_stop_mab: closing socket and stopped reading");
	}
}

static void wired_mab_activation(void *eloop_ctx, void *timeout_ctx)
{
	struct wpa_driver_wired_data *drv = eloop_ctx;
	struct hostapd_data *hapd = timeout_ctx;

	if (hapd->conf->mab_enabled) {
		wpa_printf(MSG_DEBUG, "wired_mab_activation: starting mab");
		wired_start_mab(drv);
	}
}

void wired_mab_deactivation(struct hostapd_data *hapd)
{
	struct wpa_driver_wired_data *drv = hapd->drv_priv;

	if (hapd->conf->mab_enabled) {
		if (eloop_is_timeout_registered(wired_mab_activation, drv, hapd)) {
			wpa_printf(MSG_DEBUG, "wired_mab_deactivation: cancelling mab activation mab");
			eloop_cancel_timeout(wired_mab_activation, drv, hapd);
		} else {
			if (drv->mab_sock) {
				wpa_printf(MSG_DEBUG, "wired_mab_deactivation: stopping mab");
				wired_stop_mab(drv);
				wired_handle_ebtable_rules(drv, NULL, 0);
			}
		}
	}
}

void wired_init_mab(struct hostapd_data *hapd)
{
	struct wpa_driver_wired_data *drv = hapd->drv_priv;
	struct sta_info *sta;

	/*
	 * for slow hardware, we do not want to start another
	 * timer to create a mab socket since this will reset ACL rules
	 * so we need to check for an existing mab socket
	 */
	if (drv->mab_sock)
		return;

	if (hapd->conf->mab_enabled) {
		for (sta = hapd->sta_list; sta; sta = sta->next) {
			if (sta->flags & WLAN_STA_AUTHORIZED)
				return;
		}

		if (hapd->conf->mab_activation_delay) {
			eloop_cancel_timeout(wired_mab_activation, drv, hapd);
			eloop_register_timeout(hapd->conf->mab_activation_delay, 0,
					wired_mab_activation, drv, hapd);
			wpa_printf(MSG_DEBUG, "wired_init_mab: registered timeout for mab activation");
		} else {
			wpa_printf(MSG_DEBUG, "wired_init_mab: starting mab");
			wired_start_mab(drv);
		}
	}
}

static int wired_init_interface(struct wpa_driver_wired_data *drv)
{
	struct wired_global *global = drv->global;
	struct rtnl_link *link;
	struct rtnl_link_bridge_vlan *bv;
	char *name;

	link = wpa_nl_get_link_by_family(global->cache, AF_BRIDGE,
						drv->common.ifindex);

	if (link) {
		drv->if_info.is_bridge = rtnl_link_is_bridge(link);

		drv->if_info.flags = rtnl_link_get_flags(link);
		name = rtnl_link_get_name(link);
		wpa_printf(MSG_DEBUG, "WIRED: link found in the cache for %s %d",
					name?name:"-", drv->if_info.is_bridge);

		bv = rtnl_link_bridge_get_port_vlan(link);
		if (bv)
			drv->if_info.pvid = bv->pvid;
		rtnl_link_put(link);
	} else {
		link = wpa_nl_get_link_by_family(global->cache, 0,
						 drv->common.ifindex);
		if (link) {
			drv->if_info.flags = rtnl_link_get_flags(link);
			rtnl_link_put(link);
		} else {
			wpa_printf(MSG_ERROR, "WIRED: Error: link not found in the cache for %s",
							drv->common.ifname);
		}
	}

	return 0;
}

static void * wired_driver_hapd_init(struct hostapd_data *hapd,
				     struct wpa_init_params *params)
{
	struct wpa_driver_wired_data *drv;

	drv = os_zalloc(sizeof(struct wpa_driver_wired_data));
	if (drv == NULL) {
		wpa_printf(MSG_INFO,
			   "Could not allocate memory for wired driver data");
		return NULL;
	}

	drv->global = params->global_priv;
	drv->common.ctx = hapd;

	os_strlcpy(drv->common.ifname, params->ifname,
		   sizeof(drv->common.ifname));
	drv->use_pae_group_addr = params->use_pae_group_addr;

	wpa_printf(MSG_DEBUG, "wired_driver_hapd_init called iface=%s",
			drv->common.ifname);

	if (wired_init_sockets(drv, params->own_addr)) {
		os_free(drv);
		return NULL;
	}

	wired_init_interface(drv);

	/* initialize the ebtable rules for interface */
	/* this call will be made later in hostapd_setup_interface()
	   so no need to call it now.
	*/
	if (wired_handle_ebtable_rules(drv, NULL, 0))
		return NULL;

	hapd->drv_priv = drv;
	wired_init_mab(hapd);

	dl_list_add(&drv->global->interfaces, &drv->if_list);
	return drv;
}

static void wired_driver_hapd_deinit(void *priv)
{
	struct wpa_driver_wired_data *drv = priv;
	struct hostapd_data *hapd;

	hapd = drv->common.ctx;

	wpa_printf(MSG_DEBUG, "wired_driver_hapd_deinit called iface=%s",
			drv->common.ifname);

	if (drv->common.sock >= 0) {
		eloop_unregister_read_sock(drv->common.sock);
		close(drv->common.sock);
	}

	if (drv->dhcp_sock >= 0) {
		eloop_unregister_read_sock(drv->dhcp_sock);
		close(drv->dhcp_sock);
	}

	if (drv->mab_sock >= 0) {
		eloop_unregister_read_sock(drv->mab_sock);
		close(drv->mab_sock);
	}

	/* we should remove the interface filters and any dynamic vlans if the driver is going away */
	hostapd_handle_acl_pvid_removal(hapd);

	dl_list_del(&drv->if_list);
	os_free(drv);
}

static void wired_driver_ifup(void *priv)
{
	struct wpa_driver_wired_data *drv = priv;
	int flags;

	wpa_printf(MSG_INFO, "wired_driver_ifup called");
	if (driver_wired_get_ifflags(drv->common.ifname, &flags) == 0 &&
			driver_wired_set_ifflags(drv->common.ifname, flags | IFF_UP) < 0)
		wpa_printf(MSG_WARNING, "%s: Failed to set the interface %s up",
				__func__, drv->common.ifname);
	else
		wpa_printf(MSG_DEBUG, "%s: Brought up interface %s",
				__func__, drv->common.ifname);
}

static void wired_driver_ifdown(void *priv)
{
	struct wpa_driver_wired_data *drv = priv;
	int flags;

	wpa_printf(MSG_INFO, "wired_driver_ifdown called");
	if (driver_wired_get_ifflags(drv->common.ifname, &flags) == 0 &&
			driver_wired_set_ifflags(drv->common.ifname, flags & ~IFF_UP) < 0)
		wpa_printf(MSG_WARNING, "%s: Failed to set the interface %s down",
				__func__, drv->common.ifname);
	else
		wpa_printf(MSG_DEBUG, "%s: Brought down interface %s",
				__func__, drv->common.ifname);
}

static void * wpa_driver_wired_init(void *ctx, const char *ifname)
{
	struct wpa_driver_wired_data *drv;

	drv = os_zalloc(sizeof(*drv));
	if (drv == NULL)
		return NULL;

	if (driver_wired_init_common(&drv->common, ifname, ctx) < 0) {
		os_free(drv);
		return NULL;
	}

	return drv;
}

static void wpa_driver_wired_deinit(void *priv)
{
	struct wpa_driver_wired_data *drv = priv;
	wpa_printf(MSG_INFO, "wpa_driver_wired_deinit called");

	driver_wired_deinit_common(&drv->common);
	os_free(drv);
}

static int
wired_add_fdb_entry (struct wpa_driver_wired_data *drv, const u8 *addr, int vlan)
{
	char bridge_args[254];
	struct hostapd_data *hapd = drv->common.ctx;

	if (wired_driver_hapd_is_l2_if(drv)) {
		if (drv->flags & WPA_WIRED_DATA_PVLAN_ENABLED)
			sprintf (bridge_args, "fdb add " MACSTR " dev %s vlan %d master static",
						MAC2STR(addr), drv->common.ifname, hapd->conf->pvlan_id);
		else if (vlan != -1)
			sprintf (bridge_args, "fdb add " MACSTR " dev %s vlan %d master static",
						MAC2STR(addr), drv->common.ifname, vlan);
		else
			sprintf (bridge_args, "fdb add " MACSTR " dev %s master static",
						MAC2STR(addr), drv->common.ifname);
		if (os_exec(bridge_command, bridge_args, 1)) {
			wpa_printf(MSG_ERROR, "WIRED: Error: could not add bridge fdb entry");
			return -1;
		}
		wpa_printf(MSG_DEBUG, "%s: command: %s %s", __FUNCTION__,
				bridge_command, bridge_args);
	} else {
		/* we should not be doing anything for non-L2 interfaces */
		wpa_printf(MSG_ERROR, "WIRED: Cannot add fdb entries to non-L2 Interface for %s",
				drv->common.ifname);
	}

	return 0;
}

static int
wired_del_fdb_entry (struct wpa_driver_wired_data *drv, const u8 *addr, int vlan)
{
	char bridge_args[254];
	struct hostapd_data *hapd = drv->common.ctx;

	if (wired_driver_hapd_is_l2_if(drv)) {
		if (drv->flags & WPA_WIRED_DATA_PVLAN_ENABLED)
			sprintf (bridge_args, "fdb del " MACSTR " dev %s vlan %d master static",
						MAC2STR(addr), drv->common.ifname, hapd->conf->pvlan_id);
		else if (vlan != -1)
			sprintf (bridge_args, "fdb del " MACSTR " dev %s vlan %d master static",
						MAC2STR(addr), drv->common.ifname, vlan);
		else
			sprintf (bridge_args, "fdb del " MACSTR " dev %s master static",
						MAC2STR(addr), drv->common.ifname);

		if (os_exec(bridge_command, bridge_args, 1)) {
			wpa_printf(MSG_ERROR, "WIRED: Error: could not del bridge fdb entry");
			return -1;
		}
		wpa_printf(MSG_DEBUG, "%s: command: %s %s", __FUNCTION__,
				bridge_command, bridge_args);
	} else {
		/* we should not be doing anything for non-L2 interfaces */
		wpa_printf(MSG_ERROR, "WIRED: Cannot del fdb entries to non-L2 Interface for %s",
				drv->common.ifname);
	}

	return 0;
}

static int
wired_modify_ifupdown_bridge_access_vlan(struct wpa_driver_wired_data *drv, int pvid,
		Boolean remove_only)
{
	char single_filename[64];
	FILE *iface_file = NULL;

	snprintf(single_filename, sizeof(single_filename), "/etc/network/interfaces.d/dot1x_%s.intf",
			drv->common.ifname);

	/* we may be called to clean up only */
	if (remove_only) {
		if (remove(single_filename)) {
			/* the file may or may not be there so we tried anyway */
			wpa_printf(MSG_INFO, "Did not remove bridge-access config file %s",
					single_filename);
			return -1;
		} else {
			wpa_printf(MSG_DEBUG, "removed ACL file %s",
					single_filename);
		}
		return 0;

	}
	/* For this interface, we will overwrite any existing interface dynamic vlan config file*/
	if ((pvid > 0) && drv && drv->common.ifname) {
		/* Handle ifupdown bridge access PVID assignment file for dynamic VLANs */
		iface_file = fopen(single_filename, "w");
		if (iface_file) {
			fprintf(iface_file, "# IEEE 802.1x Dynamically Generated Dynamic VLAN or Parking VLAN Configuration File\niface %s\n    bridge-access %d\n",
					drv->common.ifname, pvid);
			fclose(iface_file);
			wpa_printf(MSG_DEBUG, "Wrote bridge-access config %s", single_filename);
		} else {
			wpa_printf(MSG_ERROR, "Could not write %s", single_filename);
			return -1;
		}
	} else {
		return -1;
	}

	return 0;
}

static int
wired_change_pvid (struct wpa_driver_wired_data *drv, int old_pvid, int new_pvid)
{
	char bridge_args[254];

	wpa_printf(MSG_DEBUG, "%s: for %s old %d new %d",
			__FUNCTION__, drv->common.ifname, old_pvid, new_pvid);
	if (wired_driver_hapd_is_l2_if(drv)) {
		if (new_pvid) {
			sprintf (bridge_args, "vlan add vid %d untagged pvid dev %s",
					new_pvid, drv->common.ifname);
			if (os_exec(bridge_command, bridge_args, 1)) {
				wpa_printf(MSG_ERROR, "WIRED: Error: could not add pvid for %s",
							drv->common.ifname);
				return -1;
			}
			wpa_printf(MSG_DEBUG, "%s: command: %s %s", __FUNCTION__,
					bridge_command, bridge_args);
		}

		if (old_pvid) {
			sprintf (bridge_args, "vlan del vid %d untagged dev %s",
					old_pvid, drv->common.ifname);
			if (os_exec(bridge_command, bridge_args, 1)) {
				wpa_printf(MSG_ERROR, "WIRED: Error: could not del pvid for %s",
							drv->common.ifname);
				return -1;
			}
			wpa_printf(MSG_DEBUG, "%s: command: %s %s", __FUNCTION__,
					bridge_command, bridge_args);
		}

	} else {
		/* we should not be doing anything for non-L2 interfaces */
		wpa_printf(MSG_ERROR, "WIRED: Cannot change pvid for non-L2 Interface for %s",
				drv->common.ifname);
	}

	return 0;
}

int wired_data_pvlan_enabled(struct hostapd_data *hapd)
{
	struct wpa_driver_wired_data *drv = hapd->drv_priv;

	if (drv->flags & WPA_WIRED_DATA_PVLAN_ENABLED)
		return 1;

	return 0;
}

void wired_trigger_reauth(struct hostapd_data *hapd)
{
	struct wpa_driver_wired_data *drv = hapd->drv_priv;
	struct sta_info *sta;

	for (sta = hapd->sta_list; sta; sta = sta->next) {
		if (sta->eapol_sm)
			sta->eapol_sm->reAuthWhen = 5;
	}

	drv->flags &= ~WPA_WIRED_DATA_PVLAN_ENABLED;
}

void
wired_global_pvlan_change(void *priv, int old_pvlan_id, int new_pvlan_id)
{
	struct wired_global *global = priv;
	struct wpa_driver_wired_data *drv;
	struct hostapd_data *hapd;

	dl_list_for_each(drv, &global->interfaces,
				struct wpa_driver_wired_data, if_list) {
		hapd = drv->common.ctx;
		hapd->conf->pvlan_id = new_pvlan_id;
		if (drv->flags & WPA_WIRED_DATA_PVLAN_ENABLED) {
			wired_trigger_reauth(hapd);
		}
	}

}

static int
wired_set_sta_authorized(void *priv, const u8 *addr,
			 unsigned int total_flags, unsigned int flags_or,
			 unsigned int flags_and)
{
	struct wpa_driver_wired_data *drv = priv;
	struct hostapd_data *hapd = drv->common.ctx;
	struct sta_info *sta;

	wpa_printf(MSG_DEBUG, "WIRED: set STA entry ifname=%s ifindex=%d %s for " MACSTR,
		   drv->common.ifname,
		   drv->common.ifindex,
		   (flags_or & WPA_STA_AUTHORIZED) ? "AUTHORIZED" : "NOT AUTHORIZED",
		   MAC2STR(addr));

	if (flags_or & WPA_STA_AUTHORIZED) {
		/* loop through existing stations and allow traffic from source mac*/
		if (wired_handle_ebtable_rules(drv, NULL, 1)) {
			wpa_printf(MSG_DEBUG, "WIRED: could not write ebtable rule");
			return -1;
		}

		if (hapd && hapd->conf->pvlan_enabled &&
				(drv->flags & WPA_WIRED_DATA_PVLAN_ENABLED)) {
			wired_del_fdb_entry(drv, addr, -1);
			wired_change_pvid(drv, hapd->conf->pvlan_id, drv->if_info.pvid);
			wired_modify_ifupdown_bridge_access_vlan(drv, hapd->conf->pvlan_id, TRUE);
			drv->flags &= ~WPA_WIRED_DATA_PVLAN_ENABLED;
			hapd->auth_fail_no_sta_del = 0;
		}
		wired_stop_mab(drv);
		wired_add_fdb_entry(drv, addr, -1);
		sta = ap_get_sta(hapd, addr);
		if (sta && (sta->flags & WLAN_STA_PARKED_VLAN))
			sta->flags &= ~ WLAN_STA_PARKED_VLAN;
		if (sta && hapd && !hapd->conf->pae_detected)
			sta->flags |= WLAN_STA_MAB;
		if (sta && (sta->flags & WLAN_STA_DYNAMIC_VLAN))
			drv->flags |= WPA_WIRED_DATA_DYNAMIC_VLAN_ENABLED;
	} else {
		sta = ap_get_sta(hapd, addr);

		if (sta && sta->eapol_sm && sta->eapol_sm->eap_if &&
				sta->eapol_sm->eap_if->aaaFail ) {
			wpa_printf(MSG_DEBUG, "%s: sta address=" MACSTR " auth fail state %d",
					__FUNCTION__, MAC2STR(addr), sta->eapol_sm->eap_if->aaaFail);

			if (sta && hapd && !hapd->conf->pae_detected)
				sta->flags |= WLAN_STA_MAB;
			if ((sta->eapol_sm->authAuthFailWhileAuthenticating > 1) &&
				(drv->flags & WPA_WIRED_DATA_PVLAN_ENABLED))
				return 0;

			if (hapd && hapd->conf->pvlan_enabled && hapd->conf->pvlan_id) {
				wired_change_pvid(drv, drv->if_info.pvid, hapd->conf->pvlan_id);
				/* write a temporary config file for the new VLAN ifor ifreload */
				wired_modify_ifupdown_bridge_access_vlan(drv, hapd->conf->pvlan_id, FALSE);
				if (wired_handle_ebtable_rules(drv, sta, 1)) {
					wpa_printf(MSG_DEBUG, "WIRED: could not write ebtable rule");
					return -1;
				}
				drv->flags |= WPA_WIRED_DATA_PVLAN_ENABLED;
				sta->flags |= WLAN_STA_PARKED_VLAN;
				sta->vlan_id = hapd->conf->pvlan_id;
				hapd->auth_fail_no_sta_del = 1;
				wired_add_fdb_entry(drv, addr, -1);
			}
		} else {
			wpa_printf(MSG_DEBUG, "%s: sta address=" MACSTR " missing sm",
							__FUNCTION__, MAC2STR(addr) );
		}
	}

	return 0;
}

static int wired_sta_remove(void *priv, const u8 *addr)
{
	struct wpa_driver_wired_data *drv = priv;
	struct hostapd_data *hapd = drv->common.ctx;
	struct sta_info *sta;

	wpa_printf(MSG_DEBUG, "WIRED: wired_sta_remove called ifname=%s address=" MACSTR,
		   drv->common.ifname, MAC2STR(addr));

	sta = ap_get_sta(hapd, addr);
	wired_del_fdb_entry(drv, addr, -1);
	if (sta && hapd && !hapd->conf->pae_detected)
		sta->flags &= ~ WLAN_STA_MAB;
	if (drv->flags & WPA_WIRED_DATA_PVLAN_ENABLED) {
		wired_change_pvid(drv, hapd->conf->pvlan_id, drv->if_info.pvid);
		wired_modify_ifupdown_bridge_access_vlan(drv, drv->if_info.pvid, TRUE);
		drv->flags &= ~WPA_WIRED_DATA_PVLAN_ENABLED;
		if (sta && (sta->flags & WLAN_STA_PARKED_VLAN))
			sta->flags &= ~ WLAN_STA_PARKED_VLAN;
	}
	if (drv->flags & WPA_WIRED_DATA_DYNAMIC_VLAN_ENABLED) {
		wired_change_pvid(drv, hapd->conf->vlan->vlan_id, drv->if_info.pvid);
		wired_modify_ifupdown_bridge_access_vlan(drv, drv->if_info.pvid, TRUE);
		drv->flags &= ~WPA_WIRED_DATA_DYNAMIC_VLAN_ENABLED;
		if (sta && (sta->flags & WLAN_STA_DYNAMIC_VLAN))
			sta->flags &= ~ WLAN_STA_DYNAMIC_VLAN;
	}
	return 0;
}


static int wired_sta_deauth(void *priv, const u8 *own_addr, const u8 *addr,
			    int reason)
{
	struct wpa_driver_wired_data *drv = priv;
	struct hostapd_data *hapd = drv->common.ctx;
	struct sta_info *sta;

	sta = ap_get_sta(hapd, addr);

	wpa_printf(MSG_DEBUG, "WIRED: wired_sta_deauth called ifname=%s address=" MACSTR,
		   drv->common.ifname, MAC2STR(addr));

	if (sta && (sta->flags & WLAN_STA_PARKED_VLAN))
		sta->flags &= ~ WLAN_STA_PARKED_VLAN;
	if (sta && hapd && !hapd->conf->pae_detected)
		sta->flags &= ~ WLAN_STA_MAB;

	if (sta && (sta->flags & WLAN_STA_AUTHORIZED ||
			hapd->auth_fail_no_sta_del)) {
		if (wired_handle_ebtable_rules(drv, sta, 0)) {
			wpa_printf(MSG_DEBUG, "WIRED: could not write ebtable rule");
			return -1;
		}
		wired_del_fdb_entry(drv, addr, -1);
		if (drv->flags & WPA_WIRED_DATA_PVLAN_ENABLED) {
			wired_change_pvid(drv, hapd->conf->pvlan_id, drv->if_info.pvid);
			wired_modify_ifupdown_bridge_access_vlan(drv, drv->if_info.pvid, TRUE);
			drv->flags &= ~WPA_WIRED_DATA_PVLAN_ENABLED;
		}
		if (drv->flags & WPA_WIRED_DATA_DYNAMIC_VLAN_ENABLED) {
			wired_change_pvid(drv, hapd->conf->vlan->vlan_id, drv->if_info.pvid);
			wired_modify_ifupdown_bridge_access_vlan(drv, drv->if_info.pvid, TRUE);
			drv->flags &= ~WPA_WIRED_DATA_DYNAMIC_VLAN_ENABLED;
		}
	}
	return 0;
}

static void
wired_interface_type_change (struct wpa_driver_wired_data *drv,
				int new_type_bridge)
{
	struct hostapd_data *hapd = drv->common.ctx;
	int old_type_bridge = drv->if_info.is_bridge;

	if (old_type_bridge && !new_type_bridge) {
		wpa_printf(MSG_ERROR, "%s %s changed to L3 interface %d",
					__FUNCTION__, drv->common.ifname,
					hapd->conf->mab_enabled);
		wired_flush(drv);
		drv->if_info.is_bridge = new_type_bridge;
	} else if (!old_type_bridge && new_type_bridge) {
		wpa_printf(MSG_ERROR, "%s %s changed to bridge member",
					__FUNCTION__, drv->common.ifname);
		wired_flush(drv);
		drv->if_info.is_bridge = new_type_bridge;
		wired_init_mab(hapd);
	}
}

void wired_update_fdb_entry(struct wpa_driver_wired_data *drv)
{
	struct hostapd_data *hapd = drv->common.ctx;
	struct sta_info *sta;

	for (sta = hapd->sta_list; sta; sta = sta->next) {
		if (sta->flags & WLAN_STA_AUTHORIZED) {
			wired_del_fdb_entry(drv, sta->addr, -1);
			wired_add_fdb_entry(drv, sta->addr, -1);
		}
	}
}

void wired_del_fdb_per_vlan(struct wpa_driver_wired_data *drv, int vlan)
{
	struct hostapd_data *hapd = drv->common.ctx;
	struct sta_info *sta;

	wpa_printf(MSG_DEBUG, "%s delete for vlan %d", __FUNCTION__, vlan);
	for (sta = hapd->sta_list; sta; sta = sta->next) {
		if (sta->flags & WLAN_STA_AUTHORIZED) {
			wired_del_fdb_entry(drv, sta->addr, vlan);
		}
	}
}

static void wired_del_or_update_vlan(struct wpa_driver_wired_data *drv,
					u32 *old_bitmap, u32 *new_bitmap)
{
	Boolean update = FALSE;
	int word, bit;
	u32 mask;

	for (word = 0; word < RTNL_LINK_BRIDGE_VLAN_BITMAP_LEN; word++) {
		if (!old_bitmap[word] && !new_bitmap[word])
			continue;
		for (bit = 0; bit < 32; bit++) {
			mask = 1 << bit;
			if ((mask & new_bitmap[word]) && !(mask & old_bitmap[word])) {
				wpa_printf(MSG_DEBUG, "%s update needed for vlan %d", __FUNCTION__,
					(word * 32) + bit);
				update = TRUE;
			} else if (!(mask & new_bitmap[word]) && (mask & old_bitmap[word])) {
				wired_del_fdb_per_vlan(drv, ((word * 32) + bit));
			}
		}
	}

	if (update)
		wired_update_fdb_entry(drv);
}

static void wired_link_change_cb(struct nl_cache *cache, struct nl_object *obj,
				int action, void *data)
{
	struct rtnl_link *link = (struct rtnl_link *) obj;
	struct wired_global *global = (struct wired_global *)data;
	int ifindex;
	char *inp;
	int family;
	unsigned int flags;
	struct wpa_driver_wired_data *drv = NULL;
	struct wpa_driver_wired_data *tmp_drv;
	struct hostapd_data *hapd;

	if (!link)
		return;

	ifindex = rtnl_link_get_ifindex(link);
	family = rtnl_link_get_family(link);
	inp = rtnl_link_get_name(link);
	flags = rtnl_link_get_flags(link);

	dl_list_for_each(tmp_drv, &global->interfaces,
				struct wpa_driver_wired_data, if_list) {
		if (tmp_drv->common.ifindex == ifindex) {
			drv = tmp_drv;
			break;
		}
	}

	if (!drv)
		return;

	hapd = drv->common.ctx;

	if (action == NL_ACT_NEW) {
		wpa_printf(MSG_ERROR, "%s, family %d, ifname %s, created",
			inp, family, drv->common.ifname);

		drv->if_info.flags = flags;
	} else if (action == NL_ACT_DEL) {
		wpa_printf(MSG_DEBUG, "%s, family %d, ifname %s, deleted",
			inp, family, drv->common.ifname);
	} else if (action == NL_ACT_CHANGE) {
		struct nl_object *old_obj;
		struct rtnl_link *old_link;

		wpa_printf(MSG_ERROR, "family %d, ifname %s, flags old/new %d/%d,"
				" admin %s/%s oper %s/%s",
				family, drv->common.ifname, drv->if_info.flags, flags,
				(drv->if_info.flags & IFF_UP) ? "up" : "down",
				(flags & IFF_UP) ? "up" : "down",
				(drv->if_info.flags & IFF_RUNNING) ? "up" : "down",
				(flags & IFF_RUNNING) ? "up" : "down");

		if ((flags & IFF_RUNNING) && !(drv->if_info.flags & IFF_RUNNING)) {
			wired_init_mab(hapd);
		} else if (!(flags & IFF_RUNNING) && (drv->if_info.flags & IFF_RUNNING)) {
			wired_mab_deactivation(hapd);
			wired_flush((void *)drv);
		}

		drv->if_info.flags = flags;

		if (family == AF_BRIDGE) {
			struct rtnl_link_bridge_vlan *bv;
			struct rtnl_link_bridge_vlan *old_bv;
			int is_bridge;

			is_bridge = rtnl_link_is_bridge(link);
			wired_interface_type_change(drv, is_bridge);

			bv = rtnl_link_bridge_get_port_vlan(link);
			if (bv) {
				old_obj = nl_cache_get_last_change_orig_obj(cache);
				if (old_obj) {
					old_link = (struct rtnl_link *)old_obj;
					old_bv = rtnl_link_bridge_get_port_vlan(old_link);
					if (!old_bv) {
						wpa_printf(MSG_DEBUG, "No old BV cached");
						return;
					}

					wpa_printf(MSG_ERROR, "Bridge vlan: pvid %d:%d, %x %x\n", bv->pvid, drv->if_info.pvid,
						bv->vlan_bitmap[0], bv->untagged_bitmap[0]);
					if (drv->flags & WPA_WIRED_DATA_PVLAN_ENABLED) {
						if (bv->pvid && (hapd->conf->pvlan_id != bv->pvid)) {
							drv->if_info.pvid = bv->pvid;
							wired_change_pvid(drv, bv->pvid, hapd->conf->pvlan_id);
							wired_modify_ifupdown_bridge_access_vlan(drv, hapd->conf->pvlan_id, FALSE);
						}
					} else if (drv->flags & WPA_WIRED_DATA_DYNAMIC_VLAN_ENABLED) {
						wpa_printf(MSG_DEBUG, "Dynamic VLAN Enabled in interface");
						if (bv->pvid && hapd && (bv->pvid != hapd->conf->vlan->vlan_id)) {
							wpa_printf(MSG_DEBUG, "Changing from vlan %d to dynamic VLAN %d",
									bv->pvid, hapd->conf->vlan->vlan_id);
							drv->if_info.pvid = bv->pvid;
							wired_change_pvid(drv, bv->pvid, hapd->conf->vlan->vlan_id);
							wired_modify_ifupdown_bridge_access_vlan(drv, hapd->conf->vlan->vlan_id, FALSE);
						}
					} else {
						if ((bv->pvid != drv->if_info.pvid) ||
							os_memcmp(bv->vlan_bitmap, old_bv->vlan_bitmap,
							  RTNL_LINK_BRIDGE_VLAN_BITMAP_MAX / 8)) {
							drv->if_info.pvid = bv->pvid;
							wired_del_or_update_vlan(drv, old_bv->vlan_bitmap,
								bv->vlan_bitmap);
						}
					}

				} else {
					drv->if_info.pvid = bv->pvid;
				}

			}
		} else {
			struct rtnl_link *tmp_link;

			if (drv->if_info.is_bridge) {
				tmp_link = wpa_nl_get_link_by_family(global->cache,
						AF_BRIDGE, drv->common.ifindex);

				if (!tmp_link ||
					(tmp_link && !rtnl_link_is_bridge(tmp_link))) {
					wired_interface_type_change(drv, 0);
				}

				if (tmp_link)
					rtnl_link_put(tmp_link);
			}
		}
	}
}

static void handle_nl(int sock, void *eloop_ctx, void *sock_ctx)
{
    struct wired_global *global = eloop_ctx;
    int err;

    if ((err = nl_cache_mngr_data_ready(global->mngr)) < 0) {
        if (err == -NLE_NOMEM || err == -NLE_MSG_OVERFLOW) {
            wpa_printf(MSG_ERROR, "netlink socket buffer overflow: %s",
                    nl_geterror(err));
        } else {
            wpa_printf(MSG_ERROR, "nl_cache_mngr_data_ready failed: %s\n",
                    nl_geterror(err));
        }
    }
}

static int wired_driver_global_nl_init(struct wired_global *global)
{
	int err;

	err = nl_cache_mngr_alloc(NULL, NETLINK_ROUTE, NL_AUTO_PROVIDE,
								&global->mngr);
	if (err < 0)
		wpa_printf(MSG_ERROR, "wired: Unable to allocate cache manager: %s",
			     nl_geterror(err));

	if ((err = nl_cache_alloc_name("route/link", &global->cache)) < 0) {
            wpa_printf(MSG_ERROR, "wired: Couldn't add link cache: %s\n",
				nl_geterror(err));
	}

	/* set cache manager sync sock to 32k */
	global->sync_sock = nl_cache_mngr_get_sync_sock(global->mngr);
	nl_socket_set_msg_buf_size(global->sync_sock, 32768);

	/*
	 * Make link cache iter over all supported families. We need this for
	 * AF_BRIDGE
	 */
	nl_cache_set_flags(global->cache, NL_CACHE_AF_ITER);
	if ((err = nl_cache_mngr_add_cache(global->mngr, global->cache,
					   &wired_link_change_cb, global)) < 0) {
		wpa_printf(MSG_ERROR, "Unable to add cache route/link: %s",
			     nl_geterror(err));
	}

	if (eloop_register_read_sock(nl_cache_mngr_get_fd(global->mngr),
				handle_nl, global, NULL)) {
		wpa_printf(MSG_INFO, "Could not register nl read socket");
		return -1;
	}
	return 0;
}

static void wired_driver_global_deinit(void *priv)
{
	struct wired_global *global = priv;
	if (global == NULL)
		return;
	if (!dl_list_empty(&global->interfaces)) {
		wpa_printf(MSG_ERROR, "wired: %u interface(s) remain at "
				"wired_driver_global_deinit",
				dl_list_len(&global->interfaces));
	}

	if (global->netlink)
		netlink_deinit(global->netlink);

	eloop_unregister_read_sock(nl_cache_mngr_get_fd(global->mngr));
	nl_cache_mngr_free(global->mngr);

	os_free(global);
}

static void *wired_driver_global_init(void *ctx)
{
	struct wired_global *global;
	struct netlink_config *cfg;

	wpa_printf(MSG_ERROR, "wired: wired_driver_global_init");
	global = os_zalloc(sizeof(*global));
	if (global == NULL)
		return NULL;
	global->ctx = ctx;
	dl_list_init(&global->interfaces);

	cfg = os_zalloc(sizeof(*cfg));
	if (cfg == NULL)
		goto err;

	cfg->ctx = global;
	global->netlink = netlink_init(cfg);
	if (global->netlink == NULL) {
		os_free(cfg);
		goto err;
	}

	if (wired_driver_global_nl_init(global) < 0)
		goto err;

	return global;

err:
	wired_driver_global_deinit(global);
	return NULL;
}

static int wired_if_add(void *priv, enum wpa_driver_if_type type,
		const char *ifname, const u8 *addr,
		void *bss_ctx, void **drv_priv,
		char *force_ifname, u8 *if_addr,
		const char *bridge, int use_existing,
		int setup_ap)
{
	struct wpa_driver_wired_data *drv = priv;
	struct hostapd_data *hapd = drv->common.ctx;

	if (!wired_driver_hapd_is_l2_if(drv)) {
		/* we should not do anything for non-L2 interfaces but allow them to authenticate */
		wpa_printf(MSG_ERROR, "wired_if_add: can not handle VLAN/FDB %d for non-L2 interface %s",
				hapd->conf->vlan->vlan_id, drv->common.ifname);
		return 0;
	}

	/* first, change the PVID from default 1 to the VLAN we are using */
        if (wired_change_pvid(drv, drv->if_info.pvid, hapd->conf->vlan->vlan_id)) {
		wpa_printf(MSG_ERROR, "wired_if_add: error setting %s Dynamic VLAN %d",
				drv->common.ifname, hapd->conf->vlan->vlan_id);
		return -1;
	}
	/* Next, add a static FDB entry for the station MAC address with a VLAN*/
	if (wired_add_fdb_entry(drv, hapd->sta_list->addr, hapd->conf->vlan->vlan_id)) {
		wpa_printf(MSG_ERROR, "wired_if_add: error adding fdb entry for %s PVID %d master static",
				drv->common.ifname, hapd->conf->vlan->vlan_id);
		return -1;
	}
	wired_modify_ifupdown_bridge_access_vlan(drv, hapd->conf->vlan->vlan_id, FALSE);
	wpa_printf(MSG_DEBUG, "wired_if_add: setting %s Dynamic VLAN %d",
			drv->common.ifname, hapd->conf->vlan->vlan_id);
	return 0;
}

const struct wpa_driver_ops wpa_driver_wired_ops = {
	.name = "wired",
	.desc = "Wired Ethernet driver",
	.global_init = wired_driver_global_init,
	.global_deinit = wired_driver_global_deinit,
	.hapd_init = wired_driver_hapd_init,
	.hapd_deinit = wired_driver_hapd_deinit,
	.hapd_is_l2_if = wired_driver_hapd_is_l2_if,
	.ifdown = wired_driver_ifdown,
	.ifup = wired_driver_ifup,
	.hapd_send_eapol = wired_send_eapol,
	.sta_set_flags = wired_set_sta_authorized,
	.sta_deauth = wired_sta_deauth,
	.sta_remove = wired_sta_remove,
	.flush = wired_flush,
	.if_add = wired_if_add,
	.get_ssid = driver_wired_get_ssid,
	.get_bssid = driver_wired_get_bssid,
	.get_capa = driver_wired_get_capa,
	.init = wpa_driver_wired_init,
	.deinit = wpa_driver_wired_deinit,
};
